package be.iminds.ilabt.jfed.espec.util;

import be.iminds.ilabt.jfed.espec.model.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Consumer;

/**
 * ESpecLogger helper for managing ESpecLogListeners
 *
 * This ignores or uses an UncaughtExceptionHandler (like UncaughtGuiExceptionToBugReport) to handle exceptions in listeners,
 * instead of propagating them, as they are typically not fatal and not relevant to the method that fires the event.
 */
public class ESpecLogger {
    private static final Logger LOG = LoggerFactory.getLogger(ESpecLogger.class);

    private final List<ESpecLogListener> listeners = new CopyOnWriteArrayList<>();
    private final AtomicInteger exceptionCount = new AtomicInteger(0);

    @Nullable private final Thread.UncaughtExceptionHandler uncaughtExceptionHandler;

    public ESpecLogger() {
        this.uncaughtExceptionHandler = Thread.currentThread().getUncaughtExceptionHandler();
    }
    public ESpecLogger(@Nullable Thread.UncaughtExceptionHandler uncaughtExceptionHandler) {
        this.uncaughtExceptionHandler = uncaughtExceptionHandler;
    }

    public void addESpecLogListener(@Nonnull ESpecLogListener listener) {
        listeners.add(listener);
    }
    public void removeESpecLogListener(@Nonnull ESpecLogListener listener) {
        listeners.remove(listener);
    }

    public void firePreFileLoad() { executeSafeForAllListeners(l -> l.onPreFileLoad()); }

    public void firePostFileLoadSuccess(@Nonnull FileSource fileSource, boolean isFast, long byteSize)  {
        executeSafeForAllListeners(l -> l.onPostFileLoad(fileSource, isFast, byteSize, new ESpecLogListener.ESpecStepStatus(true)));
    }
    public void firePostFileLoadFail(@Nonnull FileSource fileSource, boolean isFast, long byteSize, @Nullable String message, @Nullable Throwable throwable)  {
        executeSafeForAllListeners(l -> l.onPostFileLoad(fileSource, isFast, byteSize, new ESpecLogListener.ESpecStepStatus(false, message, throwable)));
    }
    public void firePostFileLoadAll(boolean success)  {
        executeSafeForAllListeners(l -> l.onPostFileLoadAll(success));
    }

    public void firePreRSpec()  {
        executeSafeForAllListeners(l -> l.onPreRSpec());
    }
    public void fireRequestRSpecKnown(@Nonnull RspecSpec rspecSpec, @Nonnull String requestRSpec)  {
        executeSafeForAllListeners(l -> l.onRequestRSpecKnown(rspecSpec, requestRSpec));
    }
    /**
     * Call this when one experiment is ready (or permanently failed)
     * (Note: currently, multi rspec is not supported, so firePostRSpec is fired once, just before firePostRspecAll)
     *
     * This includes waiting for the experiment to become ready.
     * This (obviously) does not include waiting for dir upload and execute steps of ExperimentSpecification
     * This may or may not include that extra actions are ready
     *   (sharing lans,
     *     waiting for startup scripts if specified in RSpec,
     *     executing ansible as in RSpec,
     *     testing connectivity)
     * It is however recommended that this is called only when these actions are done.
     * */
    public void firePostRSpecSuccess(@Nonnull RspecSpec rspec)  {
        executeSafeForAllListeners(l -> l.onPostRSpec(rspec, new ESpecLogListener.ESpecStepStatus(true)));
    }
    public void firePostRSpecFail(@Nonnull RspecSpec rspec, @Nullable String message, @Nullable Throwable throwable)  {
        executeSafeForAllListeners(l -> l.onPostRSpec(rspec, new ESpecLogListener.ESpecStepStatus(false, message, throwable)));
    }
    /**
     * Call this when all experiments are ready (or permanently failed)
     * (Note: currently, multi rspec is not supported, so firePostRSpec is fired once,
     *  just before firePostRspecAll)
     *
     * This includes waiting for the experiment to become ready.
     * This (obviously) does not include waiting for dir upload and execute steps of ExperimentSpecification
     * This may or may not include that extra actions are ready
     *   (sharing lans,
     *     waiting for startup scripts if specified in RSpec,
     *     executing ansible as in RSpec,
     *     testing connectivity)
     * It is however recommended that this is called only when these actions are done.
     * */
    public void firePostRspecAll(boolean success, @Nonnull List<String> allNodeClientIds)  {
        executeSafeForAllListeners(l -> l.onPostRspecAll(success, allNodeClientIds));
    }

    public void firePreDir()  {
        executeSafeForAllListeners(l -> l.onPreDir());
    }
    public void firePostDirSuccess(@Nonnull DirSpec dir, @Nonnull String fullPath, @Nonnull String nodeClientId)  {
        executeSafeForAllListeners(l -> l.onPostDir(dir, fullPath, nodeClientId,
                new ESpecLogListener.ESpecStepStatus(true)));
    }
    public void firePostDirFail(@Nonnull DirSpec dir, @Nonnull String fullPath, @Nonnull String nodeClientId,
                                @Nullable String message, @Nullable Throwable throwable)  {
        executeSafeForAllListeners(l -> l.onPostDir(dir, fullPath, nodeClientId,
                new ESpecLogListener.ESpecStepStatus(false, message, throwable)));
    }
    public void firePostDirAll(boolean success)  {
        executeSafeForAllListeners(l -> l.onPostDirAll(success));
    }

    public void firePreUpload()  {
        executeSafeForAllListeners(l -> l.onPreUpload());
    }
    public void firePostUploadSuccess(@Nonnull UploadLikeSpec uploadSpec, @Nonnull String fullPath, @Nonnull String nodeClientId)  {
        executeSafeForAllListeners(l -> l.onPostUpload(uploadSpec, fullPath, nodeClientId,
                new ESpecLogListener.ESpecStepStatus(true)));
    }
    public void firePostUploadFail(@Nonnull UploadLikeSpec uploadSpec, @Nonnull String fullPath,
                                   @Nonnull String nodeClientId, @Nullable String message, @Nullable Throwable throwable)  {
        executeSafeForAllListeners(l -> l.onPostUpload(uploadSpec, fullPath, nodeClientId,
                new ESpecLogListener.ESpecStepStatus(false, message, throwable)));
    }
    public void firePostUploadAll(boolean success)  {
        executeSafeForAllListeners(l -> l.onPostUploadAll(success));
    }

    public void firePreExecute()  {
        executeSafeForAllListeners(l -> l.onPreExecute());
    }
    public void firePostExecuteSuccess(@Nonnull ExecuteSpec executeSpec,
                                @Nonnull String fullPath,
                                @Nullable String fullLogPath,
                                @Nonnull String nodeClientId,
                                int exitStatus)  {
        executeSafeForAllListeners(l -> l.onPostExecute(executeSpec, fullPath, fullLogPath, nodeClientId, exitStatus,
                new ESpecLogListener.ESpecStepStatus(true)));
    }
    public void firePostExecuteFail(@Nonnull ExecuteSpec executeSpec,
                                    @Nonnull String fullPath,
                                    @Nullable String fullLogPath,
                                    @Nonnull String nodeClientId,
                                    int exitStatus,
                                    @Nullable String message, @Nullable Throwable throwable)  {
        executeSafeForAllListeners(l -> l.onPostExecute(executeSpec, fullPath, fullLogPath, nodeClientId, exitStatus,
                new ESpecLogListener.ESpecStepStatus(false, message, throwable)));
    }
    public void firePostExecuteStepAll(@Nonnull ExecuteSpec executeSpec, boolean success)  {
        executeSafeForAllListeners(l -> l.onPostExecuteStepAll(executeSpec, success));
    }
    public void firePostExecuteAll(boolean success)  {
        executeSafeForAllListeners(l -> l.onPostExecuteAll(success));
    }

    public void firePreAnsible()  {
        executeSafeForAllListeners(l -> l.onPreAnsible());
    }
    public void firePostAnsiblePlaybookSuccess(@Nonnull AnsiblePlaybookSpec ansiblePlaybookSpec,
                                        @Nonnull String fullPath,
                                        @Nullable String fullLogPath,
                                        @Nonnull String nodeClientId)  {
        executeSafeForAllListeners(l -> l.onPostAnsiblePlaybook(ansiblePlaybookSpec, fullPath, fullLogPath, nodeClientId,
                new ESpecLogListener.ESpecStepStatus(true)));
    }
    public void firePostAnsiblePlaybookFail(@Nonnull AnsiblePlaybookSpec ansiblePlaybookSpec,
                                            @Nonnull String fullPath,
                                            @Nullable String fullLogPath,
                                            @Nonnull String nodeClientId,
                                            @Nullable String message, @Nullable Throwable throwable)  {
        executeSafeForAllListeners(l -> l.onPostAnsiblePlaybook(ansiblePlaybookSpec, fullPath, fullLogPath, nodeClientId,
                new ESpecLogListener.ESpecStepStatus(false, message, throwable)));
    }
    public void firePostAnsibleAll(boolean success)  {
        executeSafeForAllListeners(l -> l.onPostAnsibleAll(success));
    }

    private void executeSafeForAllListeners(Consumer<ESpecLogListener> c) {
        for (ESpecLogListener l : listeners) {
            try {
                c.accept(l);
            } catch (Exception e) {
                handleExceptionInListener(e);
            }
        }
    }

    /**
     * We only handle the first exception with an uncaughtExceptionhandler, if present.
     * Otherwise, the user might see too much similar exceptions.
     *
     * @param e
     */
    private void handleExceptionInListener(Exception e) {
        int count = exceptionCount.getAndIncrement();
        if (count == 0 && uncaughtExceptionHandler != null) {
            try {
                uncaughtExceptionHandler.uncaughtException(Thread.currentThread(), e);
            } catch (Exception e2) {
                LOG.error("Ignoring exception in uncaughtExceptionHandler", e2);
            }
        } else {
            LOG.error("Ignoring exception "+count+" in ESpecLogger listener", e);
        }
    }
}
